//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#pragma hdrstop

#include "IvoryDraw.h"
#include "IvoryBitmap.cpp"
#include "IvoryPalette.cpp"
#include "IvorySurface.cpp"
#include "IvoryCanvas.cpp"
#include "IvoryError.cpp"
#include "IvoryMapper.cpp"

#ifdef IVORY_smartinit
#pragma package(smart_init)
#endif

IVORY_dcr // use resource ivory.dcr
//---------------------------------------------------------------------------
TIvoryDraw* TheIvoryDraw = NULL;
   // Pointer to the Ivory Draw component. Only 1 Ivory Draw component can be
   // active at the same time.
static TList* DDSList;
   // Direct Draw Surface list
static TList* DDPList;
   // Direct Draw Palette list

static HINSTANCE hDll = NULL;
typedef HRESULT __stdcall (*PROCDirectDrawCreate)(GUID FAR* lpGUID, LPDIRECTDRAW FAR* lplpDD, IUnknown FAR* pUnkOuter);
static PROCDirectDrawCreate PDirectDrawCreate = NULL;

//---------------------------------------------------------------------------
static inline TIvoryDraw* ValidCtrCheckDraw()
{
   return new TIvoryDraw(NULL);
}
//---------------------------------------------------------------------------
TColor __fastcall ToTColor(int palindex)
{
   if(TheIvoryDraw == NULL) throw EIvoryError(iecPaletteError, "ToTColor_TheIvoryDraw");
   if(TheIvoryDraw->Palette == NULL) throw EIvoryError(iecPaletteError, "ToTColor_Palette");
   return TheIvoryDraw->Palette->Color[palindex];
}
//---------------------------------------------------------------------------
__fastcall TIvoryDraw::TIvoryDraw(Classes::TComponent* AOwner)
 : TComponent(AOwner), FAutoActivate(true), FAutoHideCursor(true),
   FActive(false), FExclusiveMode(false),
   FDirectDraw(NULL), FPrimarySurface(NULL), FPalette(NULL)
{
   FHandle = ((TWinControl*)AOwner)->Handle;
   if(TheIvoryDraw == NULL) TheIvoryDraw = this;
   Application->OnException = HandleException;
   TForm* Form = dynamic_cast<TForm*>(AOwner);
   if(Form)
   {
      FOriginalFormCreate = Form->OnCreate;
      Form->OnCreate = FormCreate;
   }
   FPrimarySurface = new TIvorySurface(this);
   FPrimarySurface->FSurfaceType = stPrimary;
   FTheBackSurface = FBackSurface = new TIvorySurface(this);
   FBackSurface->FSurfaceType = stBack;
}
//---------------------------------------------------------------------------
__fastcall TIvoryDraw::~TIvoryDraw()
{
   if(Application->OnException == HandleException)
      Application->OnException = NULL;
   Active = false;         // Back to normal Windows screen mode
   if(hDll != NULL) FreeLibrary(hDll);
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::HandleException(TObject* Sender, Exception* E)
{
   if(OnException) OnException(Sender, E);
   else
   {
      if(TheIvoryDraw) TheIvoryDraw->Active = false;
      ShowMessage(E->Message);
      Application->Terminate();
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::AfterConstruction()
{
   if(OnCreate) OnCreate(this);
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::BeforeDestruction()
{
   if(OnDestroy) OnDestroy(this);
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::AddSurface(TIvorySurface* surface)
{
   if(!Active) return;
   DDSList->Add(surface);
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::ReleaseSurface(TIvorySurface* surface)
{
   if(!Active) return;
   for(int i = 0; i < DDSList->Count; i++)
   {
      if(DDSList->Items[i] == (void*)surface)
      {
         DDSList->Delete(i);
         break;
      }
   }
   surface->Release();
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::AddPalette(TIvoryPalette* pal)
{
   if(!Active) return;
   DDPList->Add(pal);
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::ReleasePalette(TIvoryPalette* pal)
{
   if(!Active) return;
   for(int i = 0; i < DDPList->Count; i++)
   {
      if(DDPList->Items[i] == (void*)pal)
      {
         DDPList->Delete(i);
         break;
      }
   }
   pal->Release();
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::AutoCreateSurfaces()
{
   for(int i = 0; i < ComponentCount; i++)
   {
      if(dynamic_cast<TIvorySurface*>(Components[i]))
      {
         TIvorySurface* surface = (TIvorySurface*)Components[i];
         if(surface->DirectDrawSurface)
            surface->CreateOffScreen();
      }
   }
}
//---------------------------------------------------------------------------
// Activate or inactivate Ivory Draw
void __fastcall TIvoryDraw::SetActive(bool active)
{
   HRESULT ddrval;
   if(active == Active) return;
   if(active)
   {
      if(hDll == NULL)
      {
         hDll = LoadLibrary("DDRAW.DLL");
/*   if(hDll == NULL)
      throw...*/
         PDirectDrawCreate = (PROCDirectDrawCreate)GetProcAddress(hDll, "DirectDrawCreate");
/*   if(PDirectDrawCreate == NULL)
      throw...*/
      }

      TForm* form = dynamic_cast<TForm*>(Owner);
      form->OnPaint = FormPaint;
      if(FAutoHideCursor) form->Cursor = crNone;

      DDSList = new TList();
      DDPList = new TList();

      if(TheIvoryDraw && TheIvoryDraw!=this) TheIvoryDraw->Active = false;
         // Only 1 Ivory Draw can be active at the same time; we must
         // deactivate the other one before activiting this one

      TheIvoryDraw = this;

      if(FDirectDraw == NULL) // If there is no Direct Draw created, create one
      {
         ddrval = PDirectDrawCreate(NULL, &FDirectDraw, NULL);
         if(ddrval != DD_OK) throw EIvoryError(ddrval, "TIvoryDraw::SetActive_DirectDrawCreate");
      }
      if(ExclusiveMode) // if exclusive mode is enabled, change to exclusive mode
      {
         ddrval = DirectDraw->SetCooperativeLevel(FHandle, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN | DDSCL_ALLOWREBOOT);
         if(ddrval != DD_OK) throw EIvoryError(ddrval, "TIvoryDraw::SetActive_SetCooperativeLevel#1");
         ddrval = DirectDraw->SetDisplayMode(640, 480, 8);
         if(ddrval != DD_OK) throw EIvoryError(ddrval, "TIvoryDraw::SetActive_SetDisplayMode");
      }
      else
      {
         if(form)
         {
            form->Left = 0;
            form->Top = 0;
            form->ClientWidth = 640;
            form->ClientHeight = 480;
         }
         ddrval = DirectDraw->SetCooperativeLevel(FHandle, DDSCL_NORMAL);
         if(ddrval != DD_OK) throw EIvoryError(ddrval, "TIvoryDraw::SetActive_SetCooperativeLevel#2");
      }
      FActive = true;
      // this is the place where we are setting the Active bit, because now
      // the DirectDraw object is created and the cooperative mode is set,
      // so any DirectDraw function can be called; note that the primary and the
      // back surfaces are not yet set up

      if(FPrimarySurface->DirectDrawSurface == NULL) // If no primary surface is created, create it
      {
//       FPrimarySurface = new TIvorySurface(this);
         FPrimarySurface->CreatePrimary();
      }
      if(FBackSurface->DirectDrawSurface == NULL) // If no back surface is created, create it
      {
//       FBackSurface = new TIvorySurface(this);
         FBackSurface->CreateBack(FPrimarySurface);
      }
      if(Palette) // If Palette is specified, use it
      {
         TIvoryPalette* ip = Palette;
         Palette = NULL;
         Palette = ip;
      }
      BackSurface->Paint();
      AutoCreateSurfaces();

      if(OnActivate) OnActivate(this);
   }
   else
   {
      if(OnDeactivate) OnDeactivate(this);

      DirectDraw->FlipToGDISurface();
      DirectDraw->SetCooperativeLevel(FHandle, DDSCL_NORMAL);
         // Back to normal Windows screen mode

      FActive = false;
      PrimarySurface->Release(); // Release the primary surface
      FTheBackSurface->Release();      // Release the back surface
      while(DDSList->Count)
      {
         ((TIvorySurface*)DDSList->Items[0])->Release();
         DDSList->Delete(0);
      }
      delete DDSList;
      while(DDPList->Count)
      {
         ((TIvoryPalette*)DDPList->Items[0])->Release();
         DDPList->Delete(0);
      }
      delete DDPList;
      if(Palette)
         Palette->Release();  // Release the current palette (if there is one)
      if(DirectDraw)
         DirectDraw->Release();
      TheIvoryDraw = NULL;
   }
}
//---------------------------------------------------------------------------
// Set exclusive mode
void __fastcall TIvoryDraw::SetExclusiveMode(bool exclusive)
{
   if(!Active)
      FExclusiveMode = exclusive;
   // You can only set ExclusiveMode if Ivory Draw is not yet active.
}
//---------------------------------------------------------------------------
// Set and use a palette
void __fastcall TIvoryDraw::SetPalette(TIvoryPalette* palette)
{
   if(Active && palette) // If Ivory Draw is active and palette is not NULL, use the palette
   {
      HRESULT ddrval;
      if(palette->DirectDrawPalette == NULL) // if the palette is not created yet, create it
         palette->CreateFromBitmap();
      ddrval = PrimarySurface->DirectDrawSurface->SetPalette(palette->DirectDrawPalette);
      if(ExclusiveMode && ddrval != DD_OK) throw EIvoryError(ddrval, "TIvoryDraw::SetPalette_SetPrimaryPalette");
   }
   FPalette = palette; // this line bust be here!!!
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::SetBackSurface(TIvorySurface* surface)
{
   if(surface == FBackSurface) return; // no change
   if(!Active)
   {
      if(surface)
         FBackSurface = surface;
      else
         FBackSurface = FTheBackSurface;
   }
}
//---------------------------------------------------------------------------
// Copy the back surface to the screen
void __fastcall TIvoryDraw::Flip()
{
   if(!Active) return;
   if(ExclusiveMode)
   {
      while( 1 )
      {
         HRESULT ddrval;
         ddrval = PrimarySurface->DirectDrawSurface->Flip( NULL, 0 );
//       Application->ProcessMessages();
         if(ddrval == DD_OK)
            break;
         if(ddrval == DDERR_SURFACELOST)
         {
            ddrval = PrimarySurface->DirectDrawSurface->Restore();
            if(ddrval == DD_OK)
               break;
         }
         if(ddrval != DDERR_WASSTILLDRAWING)
            break;
      }
   }
   else
   {
      RECT r1, r;
      Application->ProcessMessages();
      GetWindowRect(Handle, &r);
      r1 = Rect(0, 0, 640, 480);
      Application->ProcessMessages();
      PrimarySurface->DirectDrawSurface->Blt(&r, BackSurface->DirectDrawSurface, &r1, 0, NULL);
      Application->ProcessMessages();
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::FormPaint(TObject* Sender)
{
// if(!ExclusiveMode)
   Flip();
}
//---------------------------------------------------------------------------
void __fastcall TIvoryDraw::FormCreate(TObject* Sender)
{
   if(FOriginalFormCreate) FOriginalFormCreate(Sender);
   if(FAutoActivate) Active = true;
}
//---------------------------------------------------------------------------
namespace Ivorydraw
{
   void __fastcall IVORY_package Register()
   {
      #define REGN 5
      TComponentClass classes[REGN] = { __classid(TIvoryDraw),
         __classid(TIvorySurface), __classid(TIvoryBitmap),
         __classid(TIvoryPalette), __classid(TIvoryMapper) };
      RegisterComponents("IvoryDraw", classes, REGN-1);
   }
}
//---------------------------------------------------------------------------

